import numpy as np
import core.antstimeseries as timeseries


class Normalization(timeseries.TimeSeries):
    def __init__(self):
        super(Normalization, self).__init__()

    def normalization(self, method):
        if method == 'rms':
            root_mean_square = np.sqrt(np.mean(self.samples ** 2))
            self.samples /= root_mean_square
        elif method == 'zscore':
            pass



import preparator.dirprep as dirprep
import extractor.caller as caller
import preprocessing.downsampling as downsampling
import preprocessing.filters as filters
import preprocessing.normalization as normalization
import timefrequency.wavelet as wavelet
import timefrequency.fft as fft
import postprocessing.power as post_power
import postprocessing.scale as scale
import painter.plots as plots
import circulus.cstatistics as circular_stat
import spikes.classification as classify

import os
import datetime
import pickle
import numpy as np
import core.antslogger as log


# User interface
class Ants(caller.Caller, downsampling.Downsampling, dirprep.DirPrep,
           filters.Filters, normalization.Normalization, wavelet.Wavelet, fft.FFT,
           post_power.Power, scale.Scale, plots.Plots, circular_stat.Statistics,
           classify.Classification):

    def __init__(self):
        super(Ants, self).__init__()

    @classmethod
    def make(cls, batch_size=1):
        ants = np.array([Ants() for i in np.arange(0, batch_size)])
        return ants

    @classmethod
    def save(cls, ants, path=os.path.join(os.path.expanduser('~/Documents'), 'ANTS')):
        # variables
        today = datetime.datetime.today()
        f_name = str(today.strftime('%Y_%m_%d_%H_%M')) + '_ANTS.ants'
        save_path = os.path.join(path, f_name)

        os.makedirs(path, exist_ok=True)  # make dir

        with open(save_path, "wb") as f:  # save object
            pickle.dump(len(ants), f)
            for value in ants:
                pickle.dump(value, f)
        return save_path

    @classmethod
    def load(cls, path):
        try:
            data = []
            with open(path, 'rb') as f:
                for _ in range(pickle.load(f)):
                    data.append(pickle.load(f))
            return data
        except:  # no file
            log.logger_handler.throw_error(err_code='0002', err_msg='No Files Error')
            return np.nan



import numpy as np


class TimeSeries:
    def __init__(self):
        super(TimeSeries, self).__init__()

        # variables
        self.path = str
        self.header = list
        self.sample_frequency = int
        self.samples = np.ndarray
        self.timestamps = np.ndarray
        self.time_s = np.ndarray

        self.bandpass_samples = np.ndarray

        self.waves = np.ndarray
        self.f_power = np.ndarray
        self.t_power = np.ndarray
        self.waves_freqs = np.ndarray



import os
import warnings
from joblib import Parallel, delayed, cpu_count

import numpy as np
import math
from scipy.signal.windows import dpss
from scipy.signal import detrend

import matplotlib.pyplot as plt

import core.antstimeseries as timeseries


class Multitaper(timeseries.TimeSeries):
    def __init__(self):
        super(Multitaper, self).__init__()

    def multitaper(self, **kwargs):
        # variables
        data = self.samples
        fs = kwargs.get('fs') if 'fs' in kwargs else self.sample_frequency
        frequency_range = kwargs.get('frange') if 'frange' in kwargs else [0, 200]
        time_bandwidth = kwargs.get('tbandwidth') if 'tbandwidth' in kwargs else 3
        num_tapers = kwargs.get('num_tapers') if 'num_tapers' in kwargs else 5
        window_params = kwargs.get('window_params') if 'window_params' in kwargs else [4, 1]
        min_nfft = kwargs.get('min_nfft') if 'min_nfft' in kwargs else 0
        detrend_opt = kwargs.get('detrend_opt') if 'detrend_opt' in kwargs else 'constant'
        multiprocess = kwargs.get('multiprocess') if 'multiprocess' in kwargs else True
        n_jobs = kwargs.get('n_jobs') if 'n_jobs' in kwargs else None
        weighting = kwargs.get('weighting') if 'weighting' in kwargs else 'unity'
        plot_on = kwargs.get('plot_on') if 'plot_on' in kwargs else True
        return_fig = kwargs.get('return_fig') if 'return_fig' in kwargs else False
        save_fig = kwargs.get('save_fig') if 'save_fig' in kwargs else True
        directory = kwargs.get('directory') if 'directory' in kwargs else os.path.join(os.path.dirname(os.getcwd()),
                                                                                       'figures')
        clim_scale = kwargs.get('clim_scale') if 'clim_scale' in kwargs else False
        verbose = kwargs.get('verbose') if 'verbose' in kwargs else True
        xyflip = kwargs.get('xyflip') if 'xyflip' in kwargs else False
        ax = kwargs.get('ax') if 'ax' in kwargs else None

        # pre-processing
        [data, fs, frequency_range, time_bandwidth, num_tapers,
         winsize_samples, winstep_samples, window_start,
         num_windows, nfft, detrend_opt, plot_on, verbose] = self.process_input(frequency_range, time_bandwidth,
                                                                                num_tapers, window_params, min_nfft,
                                                                                detrend_opt, plot_on, verbose)

        # set multitaper parameters
        [window_idxs, stimes, sfreqs, freq_inds] = self.process_spectrogram_params(fs, nfft, frequency_range,
                                                                                   window_start, winsize_samples)
        # split data into segments
        data_segments = data[window_idxs]

        # COMPUTE THE MULTITAPER SPECTROGRAM
        #     STEP 1: Compute DPSS tapers based on desired spectral properties
        #     STEP 2: Multiply the data segment by the DPSS Tapers
        #     STEP 3: Compute the spectrum for each tapered segment
        #     STEP 4: Take the mean of the tapered spectra

        # compute DPSS tapers (STEP 1)
        dpss_tapers, dpss_eigen = dpss(winsize_samples, time_bandwidth, num_tapers, return_ratios=True)
        dpss_eigen = np.reshape(dpss_eigen, (num_tapers, 1))

        # pre-compute weights
        if weighting == 'eigen':
            wt = dpss_eigen / num_tapers
        elif weighting == 'unity':
            wt = np.ones(num_tapers) / num_tapers
            wt = np.reshape(wt, (num_tapers, 1))  # reshape as column vector
        else:
            wt = 0

        # Set up calc_mts_segment() input arguments
        mts_params = (dpss_tapers, nfft, freq_inds, detrend_opt, num_tapers, dpss_eigen, weighting, wt)

        if multiprocess:  # use multiprocessing
            n_jobs = max(cpu_count() - 1, 1) if n_jobs is None else n_jobs
            mt_spectrogram = np.vstack(Parallel(n_jobs=n_jobs)(delayed(self.calc_mts_segment)(
                data_segments[num_window, :], *mts_params) for num_window in range(num_windows)))

        else:  # if no multiprocessing, compute normally
            mt_spectrogram = np.apply_along_axis(self.calc_mts_segment, 1, data_segments, *mts_params)

        # multiply the data segment by the DPSS tapers (STEP 2)
        # compute the spectrum for each tapered segment (STEP 3)
        # take the mean of the tapered spectra (STEP 4)
        # Compute one-sided PSD spectrum
        mt_spectrogram = mt_spectrogram.T
        dc_select = np.where(sfreqs == 0)[0]
        nyquist_select = np.where(sfreqs == fs / 2)[0]
        select = np.setdiff1d(np.arange(0, len(sfreqs)), np.concatenate((dc_select, nyquist_select)))

        mt_spectrogram = np.vstack([mt_spectrogram[dc_select, :], 2 * mt_spectrogram[select, :],
                                    mt_spectrogram[nyquist_select, :]]) / fs

        # Flip if requested
        if xyflip:
            mt_spectrogram = mt_spectrogram.T

        # Plot multitaper spectrogram
        if plot_on:
            # convert from power to dB
            spect_data = self.nanpow2db(mt_spectrogram)

            # Set x and y axes
            dx = stimes[1] - stimes[0]
            dy = sfreqs[1] - sfreqs[0]
            extent = [stimes[0] - dx, stimes[-1] + dx, sfreqs[-1] + dy, sfreqs[0] - dy]

            # Plot spectrogram
            if ax is None:
                fig, ax = plt.subplots()
            else:
                fig = ax.get_figure()
            im = ax.imshow(spect_data, extent=extent, aspect='auto')
            fig.colorbar(im, ax=ax, label='PSD (dB)', shrink=0.8)
            ax.set_xlabel("Time (HH:MM:SS)")
            ax.set_ylabel("Frequency (Hz)")
            im.set_cmap(plt.cm.get_cmap('cet_rainbow4'))
            ax.invert_yaxis()

            # Scale colormap
            if clim_scale:
                clim = np.percentile(spect_data, [5, 98])  # from 5th percentile to 98th
                im.set_clim(clim)  # actually change colorbar scale

            if return_fig:
                return mt_spectrogram, stimes, sfreqs, (fig, ax)

            if save_fig:
                fig.savefig(os.path.join(directory, 'multitaper.png'))

        return mt_spectrogram, stimes, sfreqs

    def process_input(self, frequency_range=None, time_bandwidth=5, num_tapers=None, window_params=None, min_nfft=0,
                      detrend_opt='linear', plot_on=True, verbose=True):
        """ Helper function to process multitaper_spectrogram() arguments
                Arguments:
                        data (1d np.array): time series data-- required
                        fs (float): sampling frequency in Hz  -- required
                        frequency_range (list): 1x2 list - [<min frequency>, <max frequency>] (default: [0 nyquist])
                        time_bandwidth (float): time-half bandwidth product (window duration*half bandwidth of main lobe)
                                                (default: 5 Hz*s)
                        num_tapers (int): number of DPSS tapers to use (default: None [will be computed
                                          as floor(2*time_bandwidth - 1)])
                        window_params (list): 1x2 list - [window size (seconds), step size (seconds)] (default: [5 1])
                        min_nfft (int): minimum allowable NFFT size, adds zero padding for interpolation (closest 2^x)
                                        (default: 0)
                        detrend_opt (string): detrend data window ('linear' (default), 'constant', 'off')
                                              (Default: 'linear')
                        plot_on (True): plot results (default: True)
                        verbose (True): display spectrogram properties (default: true)
                Returns:
                        data (1d np.array): same as input
                        fs (float): same as input
                        frequency_range (list): same as input or calculated from fs if not given
                        time_bandwidth (float): same as input or default if not given
                        num_tapers (int): same as input or calculated from time_bandwidth if not given
                        winsize_samples (int): number of samples in single time window
                        winstep_samples (int): number of samples in a single window step
                        window_start (1xm np.array): array of timestamps representing the beginning time for each window
                        num_windows (int): number of windows in the data
                        nfft (int): length of signal to calculate fft on
                        detrend_opt ('string'): same as input or default if not given
                        plot_on (bool): same as input
                        verbose (bool): same as input
        """

        # variables
        data = self.samples
        fs = self.sample_frequency

        # Make sure data is 1 dimensional np array
        if len(data.shape) != 1:
            if (len(data.shape) == 2) & (data.shape[1] == 1):  # if it's 2d, but can be transferred to 1d, do so
                data = np.ravel(data[:, 0])
            elif (len(data.shape) == 2) & (data.shape[0] == 1):  # if it's 2d, but can be transferred to 1d, do so
                data = np.ravel(data.T[:, 0])
            else:
                raise TypeError("Input data is the incorrect dimensions. Should be a 1d array with shape (n,) where n is \
                                the number of data points. Instead data shape was " + str(data.shape))

        # Set frequency range if not provided
        if frequency_range is None:
            frequency_range = [0, fs / 2]

        # Set detrending method
        detrend_opt = detrend_opt.lower()
        if detrend_opt != 'linear':
            if detrend_opt in ['const', 'constant']:
                detrend_opt = 'constant'
            elif detrend_opt in ['none', 'false', 'off']:
                detrend_opt = 'off'
            else:
                raise ValueError("'" + str(detrend_opt) + "' is not a valid argument for detrend_opt. The choices " +
                                 "are: 'constant', 'linear', or 'off'.")
        # Check if frequency range is valid
        if frequency_range[1] > fs / 2:
            frequency_range[1] = fs / 2
            warnings.warn('Upper frequency range greater than Nyquist, setting range to [' +
                          str(frequency_range[0]) + ', ' + str(frequency_range[1]) + ']')

        # Set number of tapers if none provided
        if num_tapers is None:
            num_tapers = math.floor(2 * time_bandwidth) - 1

        # Warn if number of tapers is suboptimal
        if num_tapers != math.floor(2 * time_bandwidth) - 1:
            warnings.warn('Number of tapers is optimal at floor(2*TW) - 1. consider using ' +
                          str(math.floor(2 * time_bandwidth) - 1))

        # If no window params provided, set to defaults
        if window_params is None:
            window_params = [5, 1]

        # Check if window size is valid, fix if not
        if window_params[0] * fs % 1 != 0:
            winsize_samples = round(window_params[0] * fs)
            warnings.warn('Window size is not divisible by sampling frequency. Adjusting window size to ' +
                          str(winsize_samples / fs) + ' seconds')
        else:
            winsize_samples = window_params[0] * fs

        # Check if window step is valid, fix if not
        if window_params[1] * fs % 1 != 0:
            winstep_samples = round(window_params[1] * fs)
            warnings.warn('Window step size is not divisible by sampling frequency. Adjusting window step size to ' +
                          str(winstep_samples / fs) + ' seconds')
        else:
            winstep_samples = window_params[1] * fs

        # Get total data length
        len_data = len(data)

        # Check if length of data is smaller than window (bad)
        if len_data < winsize_samples:
            raise ValueError("\nData length (" + str(len_data) + ") is shorter than window size (" +
                             str(winsize_samples) + "). Either increase data length or decrease window size.")

        # Find window start indices and num of windows
        window_start = np.arange(0, len_data - winsize_samples + 1, winstep_samples)
        num_windows = len(window_start)

        # Get num points in FFT
        if min_nfft == 0:  # avoid divide by zero error in np.log2(0)
            nfft = max(2 ** math.ceil(np.log2(abs(winsize_samples))), winsize_samples)
        else:
            nfft = max(max(2 ** math.ceil(np.log2(abs(winsize_samples))), winsize_samples),
                       2 ** math.ceil(np.log2(abs(min_nfft))))

        return ([data, fs, frequency_range, time_bandwidth, num_tapers,
                 int(winsize_samples), int(winstep_samples), window_start, num_windows, nfft,
                 detrend_opt, plot_on, verbose])

    def process_spectrogram_params(self, fs, nfft, frequency_range, window_start, datawin_size):
        """ Helper function to create frequency vector and window indices
                Arguments:
                     fs (float): sampling frequency in Hz  -- required
                     nfft (int): length of signal to calculate fft on -- required
                     frequency_range (list): 1x2 list - [<min frequency>, <max frequency>] -- required
                     window_start (1xm np array): array of timestamps representing the beginning time for each
                                                  window -- required
                     datawin_size (float): seconds in one window -- required
                Returns:
                    window_idxs (nxm np array): indices of timestamps for each window
                                                (nxm where n=number of windows and m=datawin_size)
                    stimes (1xt np array): array of times for the center of the spectral bins
                    sfreqs (1xf np array): array of frequency bins for the spectrogram
                    freq_inds (1d np array): boolean array of which frequencies are being analyzed in
                                              an array of frequencies from 0 to fs with steps of fs/nfft
            """

        # create frequency vector
        df = fs / nfft
        sfreqs = np.arange(0, fs, df)

        # Get frequencies for given frequency range
        freq_inds = (sfreqs >= frequency_range[0]) & (sfreqs <= frequency_range[1])
        sfreqs = sfreqs[freq_inds]

        # Compute times in the middle of each spectrum
        window_middle_samples = window_start + round(datawin_size / 2)
        stimes = window_middle_samples / fs

        # Get indexes for each window
        window_idxs = np.atleast_2d(window_start).T + np.arange(0, datawin_size, 1)
        window_idxs = window_idxs.astype(int)

        return [window_idxs, stimes, sfreqs, freq_inds]

    def calc_mts_segment(self, data_segment, dpss_tapers, nfft, freq_inds, detrend_opt, num_tapers,
                         dpss_eigen, weighting, wt):
        """ Helper function to calculate the multitaper spectrum of a single segment of data
            Arguments:
                data_segment (1d np.array): One window worth of time-series data -- required
                dpss_tapers (2d np.array): Parameters for the DPSS tapers to be used.
                                           Dimensions are (num_tapers, winsize_samples) -- required
                nfft (int): length of signal to calculate fft on -- required
                freq_inds (1d np array): boolean array of which frequencies are being analyzed in
                                          an array of frequencies from 0 to fs with steps of fs/nfft
                detrend_opt (str): detrend data window ('linear' (default), 'constant', 'off')
                num_tapers (int): number of tapers being used
                dpss_eigen (np array):
                weighting (str):
                wt (int or np array):
            Returns:
                mt_spectrum (1d np.array): spectral power for single window
        """

        # If segment has all zeros, return vector of zeros
        if all(data_segment == 0):
            ret = np.empty(sum(freq_inds))
            ret.fill(0)
            return ret

        if any(np.isnan(data_segment)):
            ret = np.empty(sum(freq_inds))
            ret.fill(np.nan)
            return ret

        # Option to detrend data to remove low frequency DC component
        if detrend_opt != 'off':
            data_segment = detrend(data_segment, type=detrend_opt)

        # Multiply data by dpss tapers (STEP 2)
        tapered_data = np.multiply(np.mat(data_segment).T, np.mat(dpss_tapers.T))

        # Compute the FFT (STEP 3)
        fft_data = np.fft.fft(tapered_data, nfft, axis=0)

        # Compute the weighted mean spectral power across tapers (STEP 4)
        spower = np.power(np.imag(fft_data), 2) + np.power(np.real(fft_data), 2)
        if weighting == 'adapt':
            # adaptive weights - for colored noise spectrum (Percival & Walden p368-370)
            tpower = np.dot(np.transpose(data_segment), (data_segment / len(data_segment)))
            spower_iter = np.mean(spower[:, 0:2], 1)
            spower_iter = spower_iter[:, np.newaxis]
            a = (1 - dpss_eigen) * tpower
            for i in range(3):  # 3 iterations only
                # Calc the MSE weights
                b = np.dot(spower_iter, np.ones((1, num_tapers))) / ((np.dot(spower_iter, np.transpose(dpss_eigen))) +
                                                                     (np.ones((nfft, 1)) * np.transpose(a)))
                # Calc new spectral estimate
                wk = (b ** 2) * np.dot(np.ones((nfft, 1)), np.transpose(dpss_eigen))
                spower_iter = np.sum((np.transpose(wk) * np.transpose(spower)), 0) / np.sum(wk, 1)
                spower_iter = spower_iter[:, np.newaxis]

            mt_spectrum = np.squeeze(spower_iter)

        else:
            # eigenvalue or uniform weights
            mt_spectrum = np.dot(spower, wt)
            mt_spectrum = np.reshape(mt_spectrum, nfft)  # reshape to 1D

        return mt_spectrum[freq_inds]

    def nanpow2db(self, y):
        """ Power to dB conversion, setting bad values to nans
            Arguments:
                y (float or array-like): power
            Returns:
                ydB (float or np array): inputs converted to dB with 0s and negatives resulting in nans
        """

        if isinstance(y, int) or isinstance(y, float):
            if y == 0:
                return np.nan
            else:
                ydB = 10 * np.log10(y)
        else:
            if isinstance(y, list):  # if list, turn into array
                y = np.asarray(y)
            y = y.astype(float)  # make sure it's a float array so we can put nans in it
            y[y == 0] = np.nan
            ydB = 10 * np.log10(y)

        return ydB



import platform


def check_os():
    os = platform.system()

    if os == 'Windows':  # Windows
        return 'win'
    elif os == 'Darwin':  # Mac
        return 'mac'
    elif os == 'Linux':  # Linux
        return 'linux'



import core.antslogger as log
import core.antstimeseries as timeseries
import postprocessing.scale as scale

import os
import numpy as np


class DirPrep(timeseries.TimeSeries):
    def __init__(self):
        super(DirPrep, self).__init__()

    @classmethod
    def get_subdirectory(cls, superior_path):
        # get directories where data is in itself
        sub_paths = [subs for subs in os.walk(superior_path) if len(subs[1]) == 0 and len(subs[-1]) != 0]
        sub_paths.sort()
        return sub_paths

    @classmethod
    def get_data_directory(cls, superior_path, **kwargs):
        # return variable
        return_dir = []

        # kwargs
        file_expander = ''  # default = all files

        if 'expander' in kwargs:
            expander = kwargs.get('expander')
            if isinstance(file_expander, str):
                file_expander = expander
            else:
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
        else:
            file_expander = ''  # all files

        # get sub-directories from superior_path
        subs = cls.get_subdirectory(superior_path=superior_path)

        for sub_i, _ in enumerate(subs):  # 1st sub-dir where data files in itself
            for file_i, _ in enumerate(subs[sub_i][-1]):  # data files in the 1st sub-dir
                if subs[sub_i][-1][file_i].endswith(file_expander):  # select only specified files
                    return_dir.append(os.path.join(subs[sub_i][0],
                                                   subs[sub_i][-1][file_i]))  # append merged directory
        return np.array(return_dir)

    @classmethod
    def get_spikes(cls, superior_path, expander='txt', **kwargs):
        # kwargs
        scale_mode = kwargs.get('scale') if 'scale' in kwargs else 'sec'
        sample_frequency = kwargs.get('sample_frequency') if 'sample_frequency' in kwargs else 2000

        spikes_path = DirPrep.get_data_directory(superior_path=superior_path, expander=expander)  # get spike time

        if scale_mode == 'sec':
            spikes = [np.loadtxt(spikes_path[i]) for i, _ in enumerate(spikes_path)]
        elif scale_mode == 'idx':
            spikes = [scale.Scale.sec_to_ts_idx(time_s=np.loadtxt(spikes_path[i]), sample_frequency=sample_frequency)
                      for i, _ in enumerate(spikes_path)]
        else:  # default
            spikes = [np.loadtxt(spikes_path[i]) for i, _ in enumerate(spikes_path)]
        return spikes



import os
import numpy as np

import core.antslogger as log
import core.antstimeseries as timeseries
import preprocessing.filters as filters
import postprocessing.scale as scale
import matplotlib.pyplot as plt


class Plots(timeseries.TimeSeries):
    def __init__(self):
        super(Plots, self).__init__()

    def power_spectrum(self, **kwargs):
        figure_path = os.path.join(os.path.dirname(os.getcwd()), 'figures')
        os.makedirs(figure_path, exist_ok=True)  # make 'figures' directory

        fig, ax = plt.subplots()

        # plot power sepctrum with sem
        if 'mean' and 'sem' and 'freqs' in kwargs:
            freqs = kwargs.get('freqs')
            mean = kwargs.get('mean')
            sem = kwargs.get('sem')
            if isinstance(freqs, np.ndarray) and isinstance(mean, np.ndarray) and isinstance(sem, np.ndarray):
                ax.fill_between(freqs, mean - sem, mean + sem, alpha=0.5)
                ax.plot(freqs, mean)
            else:
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
        else:
            ax.plot(self.waves_freqs, self.f_power)  # Default plot
            plt.autoscale(enable=True, axis='x', tight=True)
            plt.autoscale(enable=True, axis='y', tight=True)

        # X-axis (frequency) scope
        if 'xscope' in kwargs:
            frequency_scope = kwargs.get('xscope')
            if isinstance(frequency_scope, list) or isinstance(frequency_scope, np.ndarray):
                plt.xlim(frequency_scope[0], frequency_scope[-1])  # set xlim
            else:
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')

        # Y-axis (Power) scope
        if 'yscope' in kwargs:
            power_scope = kwargs.get('yscope')
            if isinstance(power_scope, list) or isinstance(power_scope, np.ndarray):
                plt.ylim(power_scope[0], power_scope[-1])  # set ylim
            else:
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')

        # Y-axis (Power) scale
        if 'yscale' in kwargs:
            y_scale_type = kwargs.get('yscale')
            if isinstance(y_scale_type, str):
                plt.yscale(y_scale_type)  # set y scale
            else:
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')

        # save spectrum
        if 'directory' in kwargs:
            directory = kwargs.get('directory')
            if isinstance(directory, str):
                fig.savefig(os.path.join(directory, 'power_spectrum.png'))
            else:
                log.logger_handler.throw_warning(warn_code='0004', warn_msg='Default Path Selected')
                fig.savefig(os.path.join(figure_path, 'power_spectrum.png'))
        else:
            fig.savefig(os.path.join(figure_path, 'power_spectrum.png'))
        return fig, os.path.join(figure_path, 'power_spectrum.png')

    def plot_eeg(self, **kwargs):
        # Variables
        samples = np.ndarray
        time_s = np.ndarray

        figure_path = os.path.join(os.path.dirname(os.getcwd()), 'figures')
        os.makedirs(figure_path, exist_ok=True)  # make 'figures' directory

        fig, ax = plt.subplots()

        # set plot samples
        if 'target_band' in kwargs:
            target_band = kwargs.get('target_band')
            if isinstance(target_band, list) or isinstance(target_band, np.ndarray):
                samples = filters.Filters.butter(samples=self.samples, sample_frequency=self.sample_frequency,
                                                 target_band=target_band)  # filtered samples
            else:
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
        else:
            samples = self.samples  # default samples

        # plot eeg
        if 'duration' in kwargs:
            duration = kwargs.get('duration')
            if isinstance(duration, list) or isinstance(duration, np.ndarray):
                time_s = scale.Scale.ts_to_sec(self.timestamps, self.sample_frequency)  # convert ts to second
                _start = int(duration[0] * self.sample_frequency)
                _end = int(duration[-1] * self.sample_frequency)
                ax.plot(time_s[_start:_end], samples[_start:_end])
            else:
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
        else:
            time_s = scale.Scale.ts_to_sec(self.timestamps, self.sample_frequency)  # convert ts to second
            _start = 0
            _end = len(time_s)
            ax.plot(time_s[_start:_end], samples[_start:_end])  # default plot
            log.logger_handler.throw_warning(warn_code='0003', warn_msg='Value Warning: Few samples can be lost.')

        # save spectrum
        if 'directory' in kwargs:
            directory = kwargs.get('directory')
            if isinstance(directory, str):
                fig.savefig(os.path.join(directory, 'eeg.png'))
            else:
                log.logger_handler.throw_warning(warn_code='0004', warn_msg='Default Path Selected')
                fig.savefig(os.path.join(figure_path, 'eeg.png'))
        else:
            fig.savefig(os.path.join(figure_path, 'eeg.png'))
        return fig, os.path.join(figure_path, 'eeg.png')



import core.antstimeseries as timeseries


class Downsampling(timeseries.TimeSeries):
    def __init__(self):
        super(Downsampling, self).__init__()

    def downsampling(self, target_fs):
        downsampling_coeff = int(self.sample_frequency / target_fs)

        self.sample_frequency = int(self.sample_frequency / downsampling_coeff)
        self.samples = self.samples[::downsampling_coeff]
        self.timestamps = self.timestamps[::downsampling_coeff]



import numpy as np
import core.antstimeseries as timeseries


class Scale(timeseries.TimeSeries):
    def __init__(self):
        super(Scale, self).__init__()

    @classmethod
    def ts_to_sec(cls, timestamps, sample_frequency):
        # variables
        _start = 0
        _duration_s = int((timestamps[-1] - timestamps[0]) / 1000000)

        time_s = np.linspace(start=_start, stop=_duration_s, num=sample_frequency * _duration_s)
        return time_s

    @classmethod
    def sec_to_ts_idx(cls, time_s, sample_frequency):
        # variables
        _start = 0

        timestamp_idx = [int(_start + (time_s[i] * sample_frequency)) for i, _ in enumerate(time_s)]
        return timestamp_idx



import core.antslogger as log
import core.antstimeseries as timeseries
import postprocessing.power as power

import scipy as sp
import numpy as np


class FFT(timeseries.TimeSeries):
    def __init__(self):
        super(FFT, self).__init__()

    def spectrogram(self, **kwargs):
        # kwargs
        samples = kwargs.get('samples') if 'samples' in kwargs else self.samples
        fs = kwargs.get('fs') if 'fs' in kwargs else self.sample_frequency
        seg = kwargs.get('nperseg') if 'nperseg' in kwargs else int(np.floor(len(samples) / 2))
        overlap = kwargs.get('noverlap') if 'noverlap' in kwargs else int(np.floor(seg / 4.5))
        window = kwargs.get('window') if 'window' in kwargs else 'hann'
        nfft = kwargs.get('nfft') if 'nfft' in kwargs else seg
        scaling = kwargs.get('scaling') if 'scaling' in kwargs else 'spectrum'
        mode = kwargs.get('mode') if 'mode' in kwargs else 'complex'

        if 'duration' in kwargs:
            duration_s = kwargs.get('duration')
            if isinstance(duration_s, list):
                duration_ts = [int((dur * fs) - 1) if dur > 0 else 0 for dur in duration_s]
                samples = samples[duration_ts[0]:duration_ts[-1]]  # slicing samples
            else:
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
                samples = samples  # Default samples
        else:
            samples = samples  # Default samples

        # spectrogram
        freqs, times, spectrum = sp.signal.spectrogram(x=samples, fs=fs, nperseg=seg, noverlap=overlap,
                                                       window=window, nfft=nfft, scaling=scaling, mode=mode)

        # save spectrum
        if 'pscale' in kwargs:
            power_scale = kwargs.get('pscale')
            if isinstance(power_scale, str):
                if power_scale == 'log':
                    self.f_power = np.log10(power.Power.to_freq_power(spectrum))
                else:
                    self.f_power = power.Power.to_freq_power(spectrum)
            else:
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
                self.f_power = power.Power.to_freq_power(spectrum)
        else:
            self.f_power = power.Power.to_freq_power(spectrum)

        self.waves = spectrum
        self.t_power = power.Power.to_time_power(spectrum)
        self.waves_freqs = freqs



import os
import numpy as np
import mat73 as readmat
import matlab
import matlab.engine as me
import core.antslogger as log
import core.antsplatform as platform
import core.antstimeseries as timeseries
import extractor.nlx_in_matlab_win as nlxW


class Caller(timeseries.TimeSeries):
    def __init__(self):
        super(Caller, self).__init__()

    def call_neuralynx(self, path):
        os = platform.check_os()

        if os == 'win':  # Windows
            # # variables
            # field_selection_flags = [1, 0, 1, 0, 1]
            # header_extraction_flag = 1
            # extraction_mode = 1
            # extraction_mode_vector = 1
            #
            # mat_engine = me.start_matlab()  # start matlab engine
            # nlx_path = str(nlxW.__path__.__dict__['_path'][0])  # get nlx_in_matlab module folder path
            # mat_engine.addpath(nlx_path)  # cd nlx_in_matlab module folder
            # try:
            #     (self.timestamps, self.sample_frequency,
            #      self.samples, self.header) = mat_engine.Nlx2MatCSC(path,
            #                                                         field_selection_flags,
            #                                                         header_extraction_flag,
            #                                                         extraction_mode,
            #                                                         extraction_mode_vector)
            # except:
            #     log.logger_handler.throw_error(err_code='0002', err_msg='No Files Error')

            ### bypass window matlab engine error; need refactoring ###
            try:
                mat = readmat.loadmat(path)  # load mat7.3 file, extracted by 'extract_to_ants.m"
                self.path = path
                self.header = list(mat['header'])
                self.sample_frequency = int(mat['sample_frequency'])
                self.samples = mat['samples']
                self.timestamps = mat['timestamps']
            except:
                log.logger_handler.throw_error(err_code='0002', err_msg='No Files Error')
        elif os == 'mac' or os == 'linux':  # Mac or linux
            try:
                mat = readmat.loadmat(path)  # load mat7.3 file, extracted by 'extract_to_ants.m"
                self.path = path
                self.header = list(mat['header'])
                self.sample_frequency = int(mat['sample_frequency'])
                self.samples = mat['samples']
                self.timestamps = mat['timestamps']
            except:
                log.logger_handler.throw_error(err_code='0002', err_msg='No Files Error')
        elif os == 'linux':  # Linux
            log.logger_handler.throw_error(err_code='0001', err_msg='OS Error')

    def call_nlx_events(self, path):
        os = platform.check_os()

        if os == 'win':
            pass
        elif os == 'mac':
            pass
        elif os == 'linux':
            pass



import numpy as np

import core.antstimeseries as timeseries


class Classification(timeseries.TimeSeries):
    def __init__(self):
        super(Classification, ).__init__()

    @classmethod
    def classify(cls, spk_s, mode='pc'):
        # variables
        duration_s = spk_s[-1] - spk_s[0]

        # firing rate
        firing_rate = len(spk_s) / duration_s

        # mad isi
        isi, mad = cls.calc_mad(spk_s=spk_s)

        # cv, cv2
        cv, cv2 = cls.calc_cv2(spk_s=spk_s)

        if mode == 'pc':  # Purkinje cell
            if firing_rate >= 40:
                if mad <= 0.008:
                    if cv2 >= 0.2:
                        return True
                else:
                    return False
            else:
                return False
        else:
            return False

    @classmethod
    def calc_mad(cls, spk_s):
        # calc ISI
        isi = np.diff(spk_s, axis=0)

        # calc MAD of the ISI
        mad = cls.mad(isi)
        return isi, mad

    @classmethod
    def mad(cls, data):
        return np.median(np.absolute(data - np.median(data, axis=0)), axis=0)

    @classmethod
    def calc_cv2(cls, spk_s):
        # pre-declaration
        cv2 = []

        # calc ISI
        isi = np.diff(spk_s, axis=0)

        # calc CV
        misi = np.mean(isi)
        stdisi = np.std(isi)
        cv = stdisi / misi

        # calc CV2
        for i, _ in enumerate(isi):
            if i != len(isi) - 1:  # before the last value
                cv2.append(np.abs(isi[i + 1] - isi[i]) / ((isi[i + 1] + isi[i]) / 2))

        cv2 = np.mean(cv2)
        return cv, cv2



import scipy as sp
import core.antstimeseries as timeseries


class Filters(timeseries.TimeSeries):
    def __init__(self):
        super(Filters, self).__init__()

    def bandpass_butter(self, target_band):
        nyquist = self.sample_frequency * 0.5
        sos = sp.signal.butter(N=6, Wn=[f / nyquist for f in target_band], btype='bandpass',
                               analog=False, output='sos')
        self.bandpass_samples = sp.signal.sosfiltfilt(sos=sos, x=self.samples)

    @classmethod
    def butter(cls, samples, sample_frequency, target_band):
        nyquist = sample_frequency * 0.5
        sos = sp.signal.butter(N=6, Wn=[f / nyquist for f in target_band], btype='bandpass',
                               analog=False, output='sos')
        bandpass_samples = sp.signal.sosfiltfilt(sos=sos, x=samples)
        return bandpass_samples



import scipy as sp
import numpy as np
import matplotlib.pyplot as plt

import core.antslogger as log
import core.antstimeseries as timeseries


class Cstatistics(timeseries.TimeSeries):

    def __init__(self):
        super(Cstatistics, self).__init__()

    def get_phase(self, **kwargs):
        if 'duration' in kwargs:  # set sample duration
            duration_s = kwargs.get('duration')
            if isinstance(duration_s, list):
                duration_ts = [int((dur * self.sample_frequency) - 1) if dur > 0 else 0 for dur in duration_s]
                samples = self.samples[duration_ts[0]:duration_ts[-1]].tolist()  # slicing samples within given duration
            else:
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
                samples = self.samples.tolist()
        else:
            samples = self.samples.tolist()

        # predict phase
        spectrum = sp.signal.hilbert(samples)  # hilbert transform
        phase_rad = np.angle(spectrum)  # convert phase to angle in radians

        return phase_rad

    @classmethod
    def phase(cls, samples, sample_frequency, **kwargs):
        if 'duration' in kwargs:  # set sample duration
            duration_s = kwargs.get('duration')
            if isinstance(duration_s, list):
                duration_ts = [int((dur * sample_frequency) - 1) if dur > 0 else 0 for dur in duration_s]
                samples = samples[duration_ts[0]:duration_ts[-1]].tolist()  # slicing samples within given duration
            else:
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
                samples = samples.tolist()
        else:
            samples = samples.tolist()

        # predict phase
        spectrum = sp.signal.hilbert(samples)  # hilbert transform
        phase_rad = np.angle(spectrum)  # convert phase to angle in radians

        return phase_rad

    @classmethod
    def mean_resultant_vector(cls, alpha, **kwargs):
        # variables
        dim = 0  # 1-D <'dim' is fixed 1-D now. However, it will revise later>
        if 'weights' in kwargs:  # when weights are given
            weights = kwargs.get('weights')
            if np.size(weights, 1) != np.size(alpha, 1) or np.size(weights, 0) != np.size(alpha, 0):
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
        else:
            weights = np.ones(np.shape(alpha))  # default weights

        # compute weighted sum of cos and sin of angles
        resultant_vector = sum(np.multiply(weights, np.exp(1j * alpha)), dim)

        # obtain angles
        resultant_vector = np.divide(np.abs(resultant_vector), sum(weights, dim))
        return resultant_vector

    @classmethod
    def kappa(cls, alpha, **kwargs):
        # variables
        alpha = np.array(alpha).reshape(-1, 1)  # to column
        N = np.shape(alpha)[-1]
        if 'weights' in kwargs:  # when weights are given
            weights = kwargs.get('weights')
            if np.size(weights, 1) > np.size(weights, 0):
                weights.reshape(-1, 1)
        else:
            weights = np.ones(np.shape(alpha))  # default weights

        if N > 1:
            R = Cstatistics.mean_resultant_vector(alpha=alpha, weights=weights)
        else:
            R = alpha

        if R < 0.53:
            kappa = 2 * R + R ** 3 + 5 * R ** 5 / 6
        elif 0.53 <= R < 0.85:
            kappa = -0.4 + 1.39 * R + 0.43 / (1 - R)
        else:
            kappa = 1 / (R ** 3 - 4 * R ** 2 + 3 * R)

        if 1 < N < 15:
            if kappa < 2:
                kappa = max(kappa - 2 * (N * kappa) ** -1, 0)
            else:
                kappa = (N - 1) ** 3 * kappa / (N ** 3 + N)
        return kappa.item()

    @classmethod
    def mean_confidence(cls, alpha, **kwargs):
        # variables
        dim = 0  # 1-D <'dim' is fixed 1-D now. However, it will revise later>
        xi = 0.05
        if 'weights' in kwargs:  # when weights are given
            weights = kwargs.get('weights')
            if np.size(weights, 1) != np.size(alpha, 1) or np.size(weights, 0) != np.size(alpha, 0):
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
        else:
            weights = np.ones(np.shape(alpha))  # default weights

        # compute ingredients for confidence limits
        r = Cstatistics.mean_resultant_vector(alpha=alpha, weights=weights)
        n = sum(weights, dim)
        R = np.multiply(n, r)
        c2 = sp.stats.chi2.ppf((1 - xi), df=1)

        # check for resultant vector length and select appropriate formula
        t = np.zeros(np.shape(r))
        shape_t = np.shape(t)  # original shape of t

        r_reshape = np.array(r).reshape(-1)  # reshape to 1-D to access element wise
        t_reshape = np.array(t).reshape(-1)  # reshape to 1-D to access element wise
        R_reshape = np.array(R).reshape(-1)  # reshape to 1-D to access element wise
        n_reshape = np.array(n).reshape(-1)  # reshape to 1-D to access element wise
        for i, _ in enumerate(np.array(r).reshape(-1)):  # reshape to 1-D to access element wise
            if r_reshape[i] < 0.9 and i > np.sqrt(c2 / 2 / n_reshape[i]):
                t_reshape[i] = np.sqrt(
                    (2 * n_reshape[i] * (2 * R_reshape[i] ** 2 - n_reshape[i] * c2)) / (4 * n_reshape[i] - c2))
            elif r_reshape[i] >= 0.9:
                t_reshape[i] = np.sqrt(
                    n_reshape[i] ** 2 - (n_reshape[i] ** 2 - R_reshape[i] ** 2) * np.exp(c2 / n_reshape[i]))
            else:
                t_reshape[i] = np.nan
                log.logger_handler.throw_warning(warn_code='0003',
                                                 warn_msg='Value Warning: Requirements for confidence levels not met.')

        # reshape to original shape
        t.reshape(shape_t)

        # apply final transform
        t = np.arccos(np.divide(t, R))
        return t

    @classmethod
    def mean_direction(cls, alpha, **kwargs):
        # variables
        dim = 0  # 1-D <'dim' is fixed 1-D now. However, it will revise later>
        if 'weights' in kwargs:
            weights = kwargs.get('weights')
            if np.size(weights, 1) != np.size(alpha, 1) or np.size(weights, 0) != np.size(alpha, 0):
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
        else:
            weights = np.ones(np.shape(alpha))  # default weights

        # compute weighted sum of cos and sin of angles
        resultant_vector = sum(np.multiply(weights, np.exp(1j * alpha)), dim)

        # obtain mean by
        mu = np.angle(resultant_vector)  # mean direction

        # confidence limits if desired
        if len(kwargs) >= 1:
            t = Cstatistics.mean_confidence(alpha=alpha, weights=weights)
            ul = mu + t  # upper 95% confidence limit
            ll = mu - t  # lower 95% confidence limit
        return dict(mean_direction=mu.item(), upper_limit=ul.item(), lower_limit=ll.item())

    @classmethod
    def von_mises_parameter(cls, alpha, **kwargs):
        # variables
        alpha = np.array(alpha).reshape(-1, 1)  # to column
        dim = 0  # 1-D <'dim' is fixed 1-D now. However, it will revise later>
        if 'weights' in kwargs:  # when weights are given
            weights = kwargs.get('weights')
        else:
            weights = np.ones(np.shape(alpha))  # default weights

        resultant_vector = Cstatistics.mean_resultant_vector(alpha=alpha, weights=weights)
        kappa = Cstatistics.kappa(alpha=resultant_vector)
        thetahat = Cstatistics.mean_direction(alpha=alpha, weights=weights)
        return thetahat, kappa

    def get_preferred_phase(self, spike_peak_ts_idx, **kwargs):
        phase_rad = Cstatistics.phase(self.samples, self.sample_frequency)  # predict phase
        spike_phase = [phase_rad[ts_idx] for ts_idx in enumerate(spike_peak_ts_idx)]  # get spike phase
        (preferred_phase, kappa) = Cstatistics.von_mises_parameter(alpha=spike_phase)  # calculate preferred phase
        return preferred_phase, kappa

    @classmethod
    def preferred_phase(cls, samples, sample_frequency, spike_peak_ts_idx, **kwargs):
        phase_rad = Cstatistics.phase(samples, sample_frequency)  # predict phase
        spike_phase = [phase_rad[ts_idx] for ts_idx in spike_peak_ts_idx]  # get spike phase
        (preferred_phase, kappa) = Cstatistics.von_mises_parameter(alpha=spike_phase)  # calculate preferred phase
        return preferred_phase, kappa



import core.antslogger as log
import numpy as np
import scipy as sp
import core.antstimeseries as timeseries


class Power(timeseries.TimeSeries):
    def __init__(self):
        super(Power, self).__init__()

    @classmethod
    def to_freq_power(cls, waves):
        f_power = np.mean((np.abs(waves)) ** 2, 1)
        return f_power

    @classmethod
    def to_time_power(cls, waves):
        t_power = np.mean((np.abs(waves)) ** 2, 0)
        return t_power

    @classmethod
    def stack_power(cls, ants):
        f_power_len = [len(ants[i].f_power) for i, _ in enumerate(ants)]
        max_len = max(f_power_len)
        freqs = ants[np.argmax(f_power_len)].waves_freqs

        # stack f_power
        stack = np.array([ants[i].f_power if f_power_len[i] == max_len else
                          np.append(ants[i].f_power, np.repeat(np.nan, max_len - f_power_len[i]))
                          for i, ant in enumerate(ants)])
        return dict(frequency=freqs, stack=stack)

    @classmethod
    def sem(cls, **kwargs):
        if 'stack' in kwargs:
            stack_dict = kwargs.get('stack')
            freqs, stack = stack_dict['frequency'], stack_dict['stack']
        elif 'batch' in kwargs:
            batch = kwargs.get('batch')
            if isinstance(batch, list) or isinstance(batch, np.ndarray):
                stack_dict = Power.stack_power(ants=batch)
                freqs, stack = stack_dict['frequency'], stack_dict['stack']
            else:
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
                return None
        else:
            log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
            return None

        mean = np.mean(stack, 0)
        se_m = sp.stats.sem(a=stack, nan_policy='omit')
        return freqs, mean, se_m



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
setuptools.setup(
    name="ants-pi",
    version="0.0.2",
    author="neuraldevops",
    author_email="minseok.kim@brain.snu.ac.kr",
    description="ANalysis for Time Series",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/minsmis/ANTS",
    project_urls={
        "Bug Tracker": "https://github.com/minsmis/ANTS/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.10",
    install_requires=[
        "mat73>=0.62",
        "matlabengine>=23.2.1",
        "matplotlib>=3.7.2",
        "numpy>=1.26.0",
        "pandas>=2.1.1",
        "pyside6>=6.6.0",
        "scikit-learn>=1.3.0",
        "scipy>=1.11.1"
    ]
)




import os
import logging


class AntsLogger:
    def __init__(self):
        super(AntsLogger, self).__init__()
        self.log_path = os.path.join(os.path.dirname(os.getcwd()), 'logs')  # log storage path
        os.makedirs(self.log_path, exist_ok=True)  # make log storage
        self.logger = logging.getLogger()  # make logger
        self.logger.setLevel(logging.INFO)  # set logging severity
        self.formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')  # print format
        self.file_handler = logging.FileHandler(os.path.join(self.log_path, 'logs.log'))  # make log file
        self.file_handler.setFormatter(self.formatter)  # set log format
        self.logger.addHandler(self.file_handler)

    def throw_error(self, err_code, err_msg):
        self.logger.error(f'%s :: %s', err_code, err_msg)

    def throw_warning(self, warn_code, warn_msg):
        self.logger.warning(f'%s :: %s', warn_code, warn_msg)


# automatic initiation
logger_handler = AntsLogger()



import os
import matlab
import matlab.engine as me
import numpy as np

import core.antslogger as log
import core.antstimeseries as timeseries
import postprocessing.power as power


class Wavelet(timeseries.TimeSeries):

    def __init__(self):
        super(Wavelet, self).__init__()
        self.PATH_MATLAB_TIMEFREQUENCY = os.path.join(os.path.dirname(os.getcwd()),
                                                      'timefrequency', 'matlab_timefrequency')

    def wavelet(self, **kwargs):
        # internal functions
        def _matlab_wavelet(cwt_samples):
            # Wavelet power spectrum - MATLAB
            matlab_engine = me.start_matlab()  # start matlab engine
            matlab_engine.cd(self.PATH_MATLAB_TIMEFREQUENCY)  # change directory to the custom matlab functions

            cwt_sampling_frequency = matlab.double(self.sample_frequency)  # type casting to matlab double

            # run cwt
            waves, freqs = matlab_engine.pyCwt(cwt_samples, cwt_sampling_frequency, nargout=2)
            return waves, freqs

        waves, freqs = [], []  # pre-declaration

        if 'duration' in kwargs:  # set sample duration
            duration_s = kwargs.get('duration')
            if isinstance(duration_s, list):
                duration_ts = [int((dur * self.sample_frequency) - 1) if dur > 0 else 0 for dur in duration_s]
                # slicing and type casting to matlab double
                cwt_samples = matlab.double(self.samples[duration_ts[0]:duration_ts[-1]].tolist())
            else:
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
                cwt_samples = matlab.double(self.samples.tolist())  # type casting to matlab double
        else:
            cwt_samples = matlab.double(self.samples.tolist())  # type casting to matlab double

        if 'operation' in kwargs:
            operation = kwargs.get('operation')
            if isinstance(operation, str):
                if operation == 'matlab':
                    waves, freqs = _matlab_wavelet(cwt_samples=cwt_samples)  # run matlab wavelet transform
                elif operation == 'python':
                    pass
                else:
                    waves, freqs = _matlab_wavelet(cwt_samples=cwt_samples)  # run matlab wavelet transform
            else:
                log.logger_handler.throw_error(err_code='0003', err_msg='Value Error')
        else:  # default wavelet
            waves, freqs = _matlab_wavelet(cwt_samples=cwt_samples)  # run matlab wavelet transform

        # save cwt results as numpy array
        self.waves = np.array(waves)
        self.waves_freqs = np.array(freqs).reshape(-1)
        self.f_power = power.Power.to_freq_power(waves=self.waves)
        self.t_power = power.Power.to_time_power(waves=self.waves)
